#!/usr/bin/env perl

use strict;
use warnings;
use Carp;

my $usage = "usage: $0 alignments.sam SS_lib_type=F,R,FR,RF\n\n";

my $sam_file = $ARGV[0] or die $usage;
my $SS_lib_type = $ARGV[1] or die $usage;

unless ($SS_lib_type =~ /^(F|R|FR|RF)$/) {
	die $usage;
}
	


main: {
	
	open (my $fh, $sam_file) or die $usage;
	
	open (my $plus_ofh, ">$sam_file.+.sam") or die "Error, cannot write to $sam_file.+.sam";
	open (my $minus_ofh, ">$sam_file.-.sam") or die "Error, cannot write to $sam_file.-.sam";
	
	while (<$fh>) {
		if (/^\@/) { next; } # header line
		my $line = $_;
		chomp;
		my @x = split(/\t/);
		my $genome_acc = $x[2];

		my $qual_string = $x[9];

		# unless ($genome_acc =~ /^700000/) { next; } ## broad-specific now, avoiding weird stuff in tophat sam output.

		my $flag = $x[1];
		
		my $query_unmapped_flag = ($flag & 0x0004) ? "qun" : "QM";
		my $pair_flag = ($flag & 0x0001) ? "PAIRED" : "notpaired";
		
        if ($query_unmapped_flag eq "qun") {
			next;
		}
		

		my $aligned_strand = ($flag & 0x0010) ? '-' : '+';
		my $opposite_strand = ($aligned_strand eq '+') ? '-' : '+';
		
		my $transcribed_strand;
		
		if ($pair_flag eq "notpaired") {
			## UNPAIRED or SINGLE READS
			unless ($SS_lib_type =~ /^(F|R)$/) {
				confess "Error, cannot have $SS_lib_type library type with unpaired reads";
			}
			
			$transcribed_strand = ($SS_lib_type eq "F") ? $aligned_strand : $opposite_strand;
		}
		else {
			
			## paired RNA-Seq reads:  left fragment is on the 3' end revcomped, and right fragment is at the 5' end sense strand.
			
			unless ($SS_lib_type =~ /^(FR|RF)$/) {
				confess "Error, cannot have $SS_lib_type library type with paired reads";
			}
			
			my $first_in_pair = ($flag & 0x0040) ? "FIRST" : "SECOND";
			
			if ($first_in_pair eq 'FIRST') {
				$transcribed_strand = ($SS_lib_type eq "FR") ? $aligned_strand : $opposite_strand;
			}
			else {
				# second pair
				$transcribed_strand = ($SS_lib_type eq "FR") ? $opposite_strand : $aligned_strand;
			}
		}
				
		my $ofh = ($transcribed_strand eq '+') ? $plus_ofh : $minus_ofh;
		
		print $ofh $line;
		
	}
	
	close $plus_ofh;
	close $minus_ofh;

	exit(0);
}


####
sub determine_first_in_pair {
	my ($qual_string) = @_;

	my @x = split(//, $qual_string);

	my $sum_left = 0;
	for (1..5) {
		my $qual = shift @x;
		$sum_left += ord($qual);
	}

	my $sum_right = 0;
	for (1..5) {
		my $qual = pop @x;
		$sum_right += ord($qual);
	}

	if ($sum_left > $sum_right) {
		return("FIRST"); # left fragment sequence has descending qualities from left to right.
	}
	elsif ($sum_left < $sum_right) {
		return("SECOND");
	}
	else {
		return("?");
	}
}

		
